// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#include "Debug.hpp"

#include <pthread.h>

#include <chrono>
#include <ctime>		 // for localtime
#include <iomanip>	 // for operator<<, setfill, setw
#include <iostream>	 // for operator<<, cerr, ostream, basic_ostream
#include <mutex>		 // for mutex, unique_lock

#include "Common.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "emper-common.h"	 // for workerid_t
#include "emper-config.h"	 // for EMPER_LOG_LEVEL
#include "io/GlobalIoContext.hpp"

using emper::io::GlobalIoContext;

static std::mutex emper_log_mutex;

void emper_add_timestamp_to(std::ostringstream& logMessage) {
	auto now = std::chrono::high_resolution_clock::now();

	auto now_time_t = std::chrono::high_resolution_clock::to_time_t(now);
	auto* now_localtime = std::localtime(&now_time_t);
	logMessage << std::put_time(now_localtime, "%H%M.");

	auto now_nanos = std::chrono::time_point_cast<std::chrono::nanoseconds>(now);
	auto time_since_epoch = now_nanos.time_since_epoch();
	long time_since_epoch_long = time_since_epoch.count();

	long remaining_nanos = time_since_epoch_long % 1000000000;
	logMessage << remaining_nanos;
}

void emper_log(const std::string& prefix, const std::string& message) {
	Worker* worker = Worker::getCurrentWorker();
	std::ostringstream logMessage;
	if (likely(worker)) {
		workerid_t workerId = worker->getWorkerId();
		std::string workerIdAsString = std::to_string(workerId);
		logMessage << std::setfill('0') << std::setw(3) << workerIdAsString << " ";
	} else {
		// Are we the global IO completer
		Runtime* runtime = Runtime::getRuntime();
		GlobalIoContext* gio = runtime ? runtime->globalIo : nullptr;
		if (gio && pthread_self() == gio->globalCompleter) {
			logMessage << "IOC ";
		} else {
			logMessage << "    ";
		}
	}

	if constexpr (emper::LOG_TIMESTAMP) {
		emper_add_timestamp_to(logMessage);
	}

	if (!prefix.empty()) {
		logMessage << " " << prefix << " ";
	} else {
		logMessage << "       ";
	}

	logMessage << message << std::endl;

	std::unique_lock<std::mutex> lock(emper_log_mutex);
	std::cerr << logMessage.str();
}

// global log_level which can be changed to control the log output
namespace emper {
enum LogLevel log_level = EMPER_LOG_LEVEL;
}
