// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#include "Emper.hpp"

#include <cassert>
#include <cerrno>
#include <cstdint>
#include <stdexcept>
#include <utility>

#include "Common.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "emper-version.h"
#include "emper.hpp"
#include "io/Future.hpp"

void async(Fiber* fiber) {
	assert(fiber != nullptr);
	Runtime* runtime = Runtime::getRuntime();
	runtime->schedule(*fiber);
}

void async(Fiber::fiber_fun_t function, void* arg) {
	Fiber* fiber = Fiber::from(std::move(function), arg);
	async(fiber);
}

void async(const Fiber::fiber_fun0_t& function) {
	Fiber* fiber = Fiber::from(function);
	async(fiber);
}

void async(Fiber::fiber_fun_t function, void* arg, workeraffinity_t* affinity) {
	Fiber* fiber = Fiber::from(std::move(function), arg, affinity);
	async(fiber);
}

void async(const Fiber::fiber_fun0_t& function, workeraffinity_t* affinity) {
	Fiber* fiber = Fiber::from(function, affinity);
	async(fiber);
}

namespace emper {

auto getFullVersion() -> std::string { return EMPER_FULL_VERSION; }

static void ensure_no_current_runtime() {
	Runtime* runtime = Runtime::getRuntime();
	if (runtime) {
		throw std::runtime_error("Runtime already initialized");
	}
}

void init_runtime() {
	ensure_no_current_runtime();

	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
	new Runtime();
	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
}

void init_runtime(workerid_t worker_count) {
	ensure_no_current_runtime();

	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
	new Runtime(worker_count);
	// NOLINTNEXTLINE(clang-analyzer-cplusplus.NewDeleteLeaks)
}

void destroy_runtime() {
	Runtime* runtime = Runtime::getRuntime();
	if (!runtime) {
		throw std::runtime_error("No runtime to destroy");
	}

	delete runtime;
}

void yield() {
	Runtime* runtime = Runtime::getRuntime();
	runtime->yield();
}

auto sleep(unsigned int seconds) -> bool {
	if constexpr (!emper::IO) {
		DIE_MSG("sleep requires emper::io");
	}

	emper::io::AlarmFuture::Timespec ts = {.tv_sec = seconds, .tv_nsec = 0};
	emper::io::AlarmFuture alarm(ts);
	int32_t res = alarm.submitAndWait();

	return res == -ETIME;
}

}	 // namespace emper
