// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#pragma once

#include <pthread.h>	// for pthread_t

#include <atomic>
#include <cassert>		 // for assert
#include <cstdint>		 // for intptr_t
#include <cstdlib>		 // for abort
#include <functional>	 // for function
#include <memory>
#include <mutex>
#include <optional>
#include <random>
#include <string>
#include <vector>	 // for vector

#include "CallerEnvironment.hpp"
#include "Context.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "NextFiberResult.hpp"
#include "Scheduler.hpp"	// for Scheduler
#include "Worker.hpp"
#include "emper-common.h"	 // for workerid_t
#include "lib/env.hpp"
#include "lib/sync/Latch.hpp"	 // for Latch
#include "lib/sync/Semaphore.hpp"
#include "sleep_strategy/WorkerSleepStrategy.hpp"

enum class LogSubsystem;
class ContextManager;
class Dispatcher;
class Fiber;
class RuntimeStrategy;
class RuntimeStrategyFactory;
class RuntimeStrategyStats;
template <typename D>
class WorkerLocalData;

namespace emper {
namespace io {
class GlobalIoContext;
class IoContext;
}	 // namespace io
namespace stats {
class Worker;
class FromAnywhere;
}	 // namespace stats
}	 // namespace emper

using emper::io::GlobalIoContext;
using emper::io::IoContext;
using emper::sleep_strategy::WorkerSleepStrategy;

class Runtime : public Logger<LogSubsystem::RUNTI> {
 private:
	static std::mutex currentRuntimeMutex;
	static Runtime* currentRuntime;

	const workerid_t workerCount;

	std::vector<std::function<void(workerid_t)>> newWorkerHooks;

	Latch workerLatch;
	Latch firstWorkerThreadExitLatch;
	Latch secondWorkerThreadExitLatch;

	RuntimeStrategy* const strategy;
	Scheduler& scheduler;
	Dispatcher& dispatcher;
	ContextManager& contextManager;
	pthread_t* threads;
	Worker** workers;
	GlobalIoContext* globalIo = nullptr;
	std::vector<IoContext*> ioContexts;
	emper::lib::sync::Semaphore ioReadySem;

	std::default_random_engine randomEngine;
	std::uniform_int_distribution<unsigned int> uniformIntDistribution;

	std::atomic<bool> terminateWorkers = false;

	bool threadsRunning = false;

	std::shared_ptr<RuntimeStrategyStats> strategyLastStats;

	WorkerLocalData<emper::stats::Worker>* workerStats;
	emper::stats::FromAnywhere* fromAnywhereStats;

	auto workerLoop(Worker* worker) -> void*;

	WorkerSleepStrategy workerSleepStrategy;

	static RuntimeStrategyFactory& DEFAULT_STRATEGY;

	static void printLastRuntimeStats();

	static auto getDefaultWorkerCount() -> workerid_t;

	static auto shouldPinWorkers() -> bool {
		return emper::lib::env::getBoolFromEnv("EMPER_PIN_WORKERS").value_or(true);
	}

 protected:
	void addNewWorkerHook(const std::function<void(workerid_t)>& hook) {
		newWorkerHooks.push_back(hook);
	};

	[[nodiscard]] inline auto getWorkerSleepStrategy() -> WorkerSleepStrategy& {
		return workerSleepStrategy;
	}

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	inline void wakeupSleepingWorkers() {
		if constexpr (::emper::WORKER_WAKEUP_STRATEGY == ::emper::WorkerWakeupStrategy::one) {
			workerSleepStrategy.notifyOne<callerEnvironment>();
		} else if constexpr (::emper::WORKER_WAKEUP_STRATEGY == ::emper::WorkerWakeupStrategy::all) {
			workerSleepStrategy.notifyAll<callerEnvironment>();
		} else {
			ABORT("Unknown CallerEnvironment");
		}
	}

	void maybeTerminateWorker() {
		if (terminateWorkers.load(std::memory_order_relaxed)) {
			LOGD("Worker terminating");
			// Switch to the original stack, which will simply terminate the thread.
			Context::switchToOriginalStack();
		}
	}

	void dispatchLoopSleep() { workerSleepStrategy.sleep(); }

 public:
	Runtime() : Runtime(getDefaultWorkerCount()) {}

	Runtime(workerid_t workerCount) : Runtime(workerCount, DEFAULT_STRATEGY) {}

	Runtime(RuntimeStrategyFactory& strategyFactory)
			: Runtime(getDefaultWorkerCount(), strategyFactory) {}

	Runtime(workerid_t workerCount, RuntimeStrategyFactory& strategyFactory,
					unsigned int seed = std::random_device()());

	~Runtime();

	inline void schedule(Fiber& fiber) {
		// Calling schedule() only works from within the EMPER runtime.
		assert(inRuntime());

		scheduler.schedule(fiber);
	}

	inline void schedule(Fiber** fibers, unsigned count) {
		// Calling schedule() only works from within the EMPER runtime.
		assert(inRuntime());

		scheduler.schedule(fibers, count);
	}

	inline void scheduleFromAnywhere(Fiber& fiber) { scheduler.scheduleFromAnywhere(fiber); }

	inline void scheduleFromAnywhere(Fiber** fibers, unsigned count) {
		scheduler.scheduleFromAnywhere(fibers, count);
	}

	void yield();

	// TODO: This should probably not be a public method of Runtime.
	auto nextFiber() -> NextFiberResult;

	// https://stackoverflow.com/a/3747462/194894
	static inline auto rand() -> int { return Worker::rand(); }

	static inline auto getWorkerId() -> workerid_t { return Worker::getCurrentWorkerId(); }

	[[nodiscard]] inline auto getWorkerCount() const -> workerid_t { return workerCount; }

	static inline auto getRuntime() -> Runtime* { return currentRuntime; }

	inline auto getContextManager() -> ContextManager& { return contextManager; }

	inline auto getStrategy() -> RuntimeStrategy& { return *strategy; }

	[[nodiscard]] inline auto isTerminating() const -> bool { return terminateWorkers; }

	void initiateTermination();

	void waitUntilFinished();

	void initiateAndWaitUntilTermination();

	void printStats(bool detailed = false);

	static auto inRuntime() -> bool { return Worker::isWorkerThread(); }

	void executeAndWait(std::function<void()> f);

	template <LogSubsystem>
	friend class Blockable;
	friend ContextManager;
	friend Scheduler;
	friend Dispatcher;
	friend GlobalIoContext;
	friend IoContext;
	template <typename T, intptr_t WS_QUEUE_SIZE, size_t WORKER_EXCLUSIVE_QUEUE_SIZE>
	friend class MemoryManager;
	template <typename>
	friend class WorkerLocalData;

	friend void emper_log(const std::string& prefix, const std::string& message);
};
