// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <mutex>
#include <queue>

#include "Blockable.hpp"
#include "CallerEnvironment.hpp"
#include "Context.hpp"

template <typename T>
class UnboundedBlockingMpscQueue : public Blockable<LogSubsystem::U_B_MPSC_Q> {
 private:
	std::atomic<Context*> blockedContext = nullptr;

	bool tPopped;
	T t;

	std::queue<T> mpscQueue;
	std::mutex queueMutex;

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void tryToWakeupBlockedContext() {
		Context* context = blockedContext.exchange(nullptr);
		if (!context) return;

		unblock<callerEnvironment>(context);
	}

	void tryToGetElement(const std::function<void(void)>& postRetrieve) {
		// tPopped indicates that 't' is a popped and usable value.
		// Therefore if we tryToGet a new 't' while 't' is available the current 't'
		// would be overridden and dropped.
		assert(!tPopped);

		std::lock_guard<std::mutex> lock(queueMutex);
		if (mpscQueue.empty()) return;

		t = mpscQueue.front();
		mpscQueue.pop();
		tPopped = true;
		postRetrieve();
	}

 public:
	UnboundedBlockingMpscQueue(Runtime& runtime) : Blockable(runtime) {}

	template <CallerEnvironment callerEnvironment = CallerEnvironment::EMPER>
	void put(T t) {
		{
			std::lock_guard<std::mutex> lock(queueMutex);
			mpscQueue.push(t);
		}

		// Micro optimization, see if there is a blocked context
		// before performing the atomic exchange operation.
		if (blockedContext != nullptr) {
			tryToWakeupBlockedContext<callerEnvironment>();
		}
	}

	void putFromAnywhere(T t) { put<CallerEnvironment::ANYWHERE>(t); }

	auto get(const std::function<void(void)>& postRetrieve) -> T {
		tPopped = false;
		tryToGetElement(postRetrieve);

		if (!tPopped) {
			Context* context = Context::getCurrentContext();
		block_again:
			block([this, context, postRetrieve] {
				tryToGetElement(postRetrieve);

				// N.B. It is important that we set blockedContext *after* we
				// called tryToGetElement(), as otherwise, multiple threads
				// may call tryToGetElement() which results in messages put in
				// the queue getting lost.
				blockedContext = context;

				if (tPopped) {
					tryToWakeupBlockedContext();
				}
			});
			if (!tPopped) {
				std::unique_lock<std::mutex> lock(queueMutex);
				// If we observed tPopped to be false without holding the
				// queueMutex, then the same must be true when holding the
				// mutex.
				assert(!tPopped);

				// Handle spurious wake-ups caused by the non atomic push and wake-up pair
				// in put(T).
				// Between the push and the wake-up a successful get() can consume the pushed
				// element and a second get() could already be blocked because the queue is now empty.
				// This blocked context is then unblocked  by the put() but the queue is empty.
				// When encountering an empty Queue after a wake-up we will block again.

				// TODO: It could be interesting to investigate if immediately retrying to
				// pop using tryToGetElement (line 67) is generally better than to immediately block
				// again (line 71) which causes a tryToGetElement call in the block hook.
				if (unlikely(mpscQueue.empty())) {
					lock.unlock();
					goto block_again;
				}

				assert(!mpscQueue.empty());

				t = mpscQueue.front();
				mpscQueue.pop();
				postRetrieve();
			}
		}
		return t;
	}

	auto size() -> size_t {
		std::lock_guard<std::mutex> lock(queueMutex);
		return mpscQueue.size();
	}
};
