// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "io/GlobalIoContext.hpp"

#include <liburing.h>
#include <liburing/io_uring.h>
#include <unistd.h>

#include <cassert>
#include <cerrno>
#include <cstdio>
#include <mutex>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Emper.hpp"
#include "Runtime.hpp"
#include "Worker.hpp"
#include "io/Future.hpp"
#include "io/IoContext.hpp"
#include "lib/TaggedPtr.hpp"
#include "lib/sync/Semaphore.hpp"
#include "sleep_strategy/WorkerSleepStrategy.hpp"

using emper::lib::TaggedPtr;

namespace emper::io {

auto GlobalIoContext::globalCompleterFunc(void* arg) -> void* {
	auto* globalIoContext = reinterpret_cast<GlobalIoContext*>(arg);

	// wait till all workers registered their IoContext's eventfd
	for (workerid_t i = 0; i < globalIoContext->runtime.workerCount; ++i) {
		globalIoContext->runtime.ioReadySem.wait();
	}

	globalIoContext->logD("submit all worker io_uring eventfds");

	// submit all eventfds in the SQ inserted by IoContext::submit_efd calls
	int submitted = io_uring_submit(&globalIoContext->ring);
	if (unlikely(submitted < 0)) {
		DIE_MSG_ERRNO("initial global io_uring submit failed");
	}

	// We have submitted all eventfds
	assert(submitted == globalIoContext->runtime.getWorkerCount());

	// Submit the read for our termination eventfd
	struct io_uring_sqe* sqe = io_uring_get_sqe(&globalIoContext->ring);
	assert(sqe);

	// We reuse the notificationEventFd member of IoContext to receive a the termination notification
	io_uring_prep_read(sqe, globalIoContext->notificationEventFd,
										 &globalIoContext->notificationEventFdBuf,
										 sizeof(globalIoContext->notificationEventFdBuf), 0);
	TaggedPtr terminationEvent(static_cast<void*>(nullptr),
														 static_cast<uint16_t>(PointerTags::TerminationEvent));
	io_uring_sqe_set_data(sqe, terminationEvent);

	submitted = io_uring_submit(&globalIoContext->ring);
	if (unlikely(submitted < 0)) {
		DIE_MSG_ERRNO("submitting termination eventfd read failed");
	}
	assert(submitted == 1);

	globalIoContext->logD("start global completer loop");
	while (true) {
		struct io_uring_cqe* cqe;

		// wait for completions
		int err = io_uring_wait_cqe(&globalIoContext->ring, &cqe);
		if (unlikely(err)) {
			if (err == -EINTR) {
				continue;
			}
			errno = -err;
			perror("io_uring_wait_cqe");
		}

		TaggedPtr tptr(io_uring_cqe_get_data(cqe));
		auto tag = static_cast<PointerTags>(tptr.getTag());

		// The cqe is for a IoContext.notificationEventFd read
		//  -> there are completions on this worker IoContext
		switch (tag) {
			// clang-11 does not support [[likely]] yet
			// TODO: remove the preprocessor check if clang has [[likely]] support
#if defined __has_attribute
#if __has_attribute(likely)
			[[likely]]	// NOLINT(clang-diagnostic-unknown-attributes)
#endif
#endif
			case PointerTags::IoContext : {
				auto* worker_io = tptr.getPtr<IoContext>();
				assert(worker_io);

				// re-add the eventfd read
				globalIoContext->prepareWorkerNotification(*worker_io);
				submitted = io_uring_submit(&globalIoContext->ring);

				if (unlikely(submitted < 0)) {
					errno = -submitted;
					DIE_MSG_ERRNO("re-submitting eventfd read to global_ring failed");
				}

				assert(submitted == 1);

				if constexpr (emper::IO_COMPLETER_BEHAVIOR == emper::IoCompleterBehavior::wakeup) {
					// Notify the worker which's CQ has a new cqe
					workerid_t worker = worker_io->worker->getWorkerId();
					globalIoContext->runtime.workerSleepStrategy.notifySpecific<CallerEnvironment::ANYWHERE>(
							worker);
				} else {
					worker_io->reapAndScheduleCompletions<CallerEnvironment::ANYWHERE>();
				}
			}
			break;

			// The cqe is for a completed Future.
			// This happens in case the global completer is dealing with a
			// partial completion.
			case PointerTags::Future: {
				auto* future = tptr.getPtr<Future>();

				int32_t res = cqe->res;

				future->recordCompletion(globalIoContext->stats, res);
				future->completeFromAnywhere(res);
			} break;

				// someone initiated termination
			case PointerTags::TerminationEvent: {
				pthread_exit(nullptr);
			} break;

			default:
				DIE_MSG("Unknown GlobalIoContext::PointerTag: " << tptr.getTag());
				break;
		}

		io_uring_cqe_seen(&globalIoContext->ring, cqe);
	}
}

void GlobalIoContext::startGlobalCompleter() {
	errno = pthread_create(&globalCompleter, nullptr, globalCompleterFunc, this);
	if (unlikely(errno)) {
		DIE_MSG_ERRNO("Creating global completer thread failed");
	}
}

void GlobalIoContext::initiateTermination() {
	uint64_t write_buf = 1;
	ssize_t bytes_written = write(notificationEventFd, &write_buf, sizeof(write_buf));
	if (bytes_written != sizeof(write_buf)) {
		DIE_MSG_ERRNO("Writing to globalIo termination eventfd failed");
	}
}

void GlobalIoContext::waitUntilFinished() const {
	errno = pthread_join(globalCompleter, nullptr);
	if (unlikely(errno)) {
		DIE_MSG_ERRNO("pthread_join failed for the globalCompleter");
	}
}

// This function must not be executed in parallel because it does not synchronize
// the SQ
void GlobalIoContext::registerWorkerIo(IoContext& workerIo) {
	// Register the worker's notificationEventFd with its io_uring
	if (unlikely(io_uring_register_eventfd(&workerIo.ring, workerIo.notificationEventFd) < 0)) {
		DIE_MSG_ERRNO("io_uring_register_eventfd failed");
	}

	std::lock_guard<std::mutex> lock(registerLock);
	prepareWorkerNotification(workerIo);
	// The sqe we prepared will be submitted to io_uring when the globalCompleter starts.
}

void GlobalIoContext::prepareWorkerNotification(IoContext& workerIo) {
	struct io_uring_sqe* sqe = io_uring_get_sqe(&ring);
	// We initialized the global ring big enough we must always get a free sqe
	assert(sqe);

	io_uring_prep_read(sqe, workerIo.notificationEventFd, &workerIo.notificationEventFdBuf,
										 sizeof(workerIo.notificationEventFdBuf), 0);
	io_uring_sqe_set_data(sqe, TaggedPtr(&workerIo, static_cast<uint16_t>(PointerTags::IoContext)));
}
}	 // namespace emper::io
