// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer, Florian Schmaus
#pragma once
#include <bits/types/struct_iovec.h>	// for iovec
#include <sys/types.h>								// for ssize_t

#include <boost/circular_buffer.hpp>
#include <boost/core/pointer_traits.hpp>
#include <chrono>		 // for nanoseconds
#include <cstddef>	 // for size_t
#include <cstdint>	 // for uint64_t, int32_t, uint32_t
#include <cstdlib>	 // for abort
#include <iostream>	 // for operator<<, basic_ostream, basi...
#include <map>			 // for map, map<>::value_compare
#include <vector>

#include "CallerEnvironment.hpp"
#include "Debug.hpp"				 // for LOGW
#include "Emper.hpp"				 // for STATS
#include "emper-common.h"		 // for workerid_t
#include "io/Future.hpp"		 // for PartialCompletableFuture, Write...
#include "io/Operation.hpp"	 // for Operation, Operation::RECV, Ope...
#include "lib/math.hpp"

class Runtime;	// lines 28-28
namespace emper {
namespace io {
class IoContext;
}
}	 // namespace emper

namespace math = emper::lib::math;

#define RETURN_IF_NO_STATS()     \
	if constexpr (!emper::STATS) { \
		return;                      \
	}

namespace emper::io {

using emper::io::Operation;
using std::chrono::nanoseconds;

class Stats {
	friend class ::Runtime;
	friend class IoContext;
	friend class Future;
	friend class PartialCompletableFuture;
	friend class SendFuture;
	friend class RecvFuture;
	friend class WritevFuture;

	static const int GLOBAL_COMPLETER_ID = -1;
	static const int AVG_ID = -2;

	int workerId = 0;
	static_assert(sizeof(int) > sizeof(workerid_t));

	enum CompletionType {
		FullCompletion = 0,
		ErrorCompletion,
		PartialCompletion,
		PartialResubmission,
		IncrementalCompletion,
		IncrementalError,
		END_OF_COMPLETION_TYPES
	};

	/**
	 * @brief Maps containing all completion events
	 *
	 * The first dimension are the different Operations defined by emper::io::Operation
	 *
	 * The second dimension is used to differentiate the available completion types per Operation
	 *   0: full completions
	 *   1: error completions
	 *   2: partial completions
	 *   3: incremental completions
	 *   4: incremental error
	 *   5: partial resubmissions
	 *   6: immediate resubmissions
	 */

	using CompletionMap = std::map<Operation, std::map<CompletionType, uint64_t>>;

	CompletionMap io_uring_completions = {
			{Operation::SEND,
			 {{FullCompletion, 0},
				{ErrorCompletion, 0},
				{PartialCompletion, 0},
				{IncrementalCompletion, 0},
				{IncrementalError, 0},
				{PartialResubmission, 0}}},
			{Operation::RECV, {{FullCompletion, 0}, {ErrorCompletion, 0}, {PartialCompletion, 0}}},
			{Operation::CONNECT, {{FullCompletion, 0}, {ErrorCompletion, 0}}},
			{Operation::ACCEPT, {{FullCompletion, 0}, {ErrorCompletion, 0}}},
			{Operation::READ,
			 {{FullCompletion, 0},
				{ErrorCompletion, 0},
				{PartialCompletion, 0},
				{IncrementalCompletion, 0},
				{IncrementalError, 0},
				{PartialResubmission, 0}}},
			{Operation::WRITE,
			 {{FullCompletion, 0},
				{ErrorCompletion, 0},
				{PartialCompletion, 0},
				{IncrementalCompletion, 0},
				{IncrementalError, 0},
				{PartialResubmission, 0}}},
			{Operation::WRITEV, {{FullCompletion, 0}, {ErrorCompletion, 0}, {PartialCompletion, 0}}},
			{Operation::CLOSE, {{FullCompletion, 0}, {ErrorCompletion, 0}}},
			{Operation::LINK_TIMEOUT, {{FullCompletion, 0}, {ErrorCompletion, 0}}},
			{Operation::TIMEOUT, {{FullCompletion, 0}, {ErrorCompletion, 0}}},
			{Operation::CANCEL, {{FullCompletion, 0}, {ErrorCompletion, 0}}},
	};
	uint64_t io_submit_full_cq = 0;
	uint64_t io_submit_full_cq_running_mean = 0;

	uint64_t completer_reap = 0;
	uint64_t completer_reaped_completions = 0;

	uint64_t worker_reap = 0;
	uint64_t worker_reaped_completions = 0;

	math::RunningAverage<float> reReapCount_average;
	unsigned reReapCount_max = 0;
	boost::circular_buffer<unsigned> reReapCount_last;

	static void printStats(IoContext* globalIoContext,
												 const std::vector<IoContext*>& workerIoContexts);

	Stats() : reReapCount_last(10) {}

	void recordCompletion(const Future& f, const int32_t& res) {
		record_completion(f.op, res, PartialCompletableFuture::DISABLE_PARTIAL_COMPLETION, f.len);
	}

	void recordCompletion(const PartialCompletableFuture& f, const int32_t& res) {
		record_completion(f.op, res, f.partialCompletion, f.len);
	}

	void recordCompletion(const WritevFuture& f, const int32_t& res) {
		size_t exp = 0;
		const auto* iov = reinterpret_cast<const struct iovec*>(f.buf);
		for (unsigned i = 0; i < f.len; ++i) {
			exp += iov[i].iov_len;
		}

		record_completion(Operation::WRITEV, res, PartialCompletableFuture::DISABLE_PARTIAL_COMPLETION,
											exp);
	}

	void record_completion(Operation op, int32_t res, ssize_t partial_completion, size_t exp) {
		auto& operation_map = io_uring_completions[op];

		// Persistent Error
		if (res < 0) {
			if (partial_completion > 0) {
				operation_map[IncrementalError]++;
			} else {
				operation_map[ErrorCompletion]++;
			}
			return;
		}

		// Make sure exp which is passed as future.len is only > 0 for Futures with a buffer
		switch (op) {
			case Operation::SEND:
			case Operation::RECV:
			case Operation::READ:
			case Operation::WRITE:
			case Operation::WRITEV:
				break;
			default:
				exp = 0;
				break;
		}

		// Full completion
		if (!exp || (uint32_t)res == exp) {
			operation_map[FullCompletion]++;
			return;
		}

		// we expect partial completion
		if (partial_completion != PartialCompletableFuture::DISABLE_PARTIAL_COMPLETION) {
			if ((size_t)(res + partial_completion) < exp) {
				operation_map[PartialResubmission]++;
			} else {
				operation_map[IncrementalCompletion]++;
			}
			return;
		}

		if ((uint32_t)res < exp) {
			operation_map[PartialCompletion]++;
			return;
		}

		LOGW("Unexpected completion case in IoStats");
		LOGW(op << ", " << res << ", " << partial_completion << ", " << exp);
		abort();
	}

	// running mean calculation taken from
	// https://math.stackexchange.com/questions/106700/incremental-averageing
	inline void record_io_submit_full_cq(nanoseconds ns) {
		RETURN_IF_NO_STATS();
		io_submit_full_cq++;

		math::calcRunningAvg(io_submit_full_cq_running_mean, io_submit_full_cq, ns.count());
	}

	template <CallerEnvironment callerEnvironment>
	inline void record_reaps(unsigned count) {
		RETURN_IF_NO_STATS();

		if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
			worker_reap++;
			worker_reaped_completions += count;
		} else {
			completer_reap++;
			completer_reaped_completions += count;
		}
	}

	void record_reReapCount(unsigned reReapCount) {
		RETURN_IF_NO_STATS();

		if (!reReapCount) {
			return;
		}

		reReapCount_average.update(reReapCount);

		if (reReapCount > reReapCount_max) {
			reReapCount_max = reReapCount;
		}

		reReapCount_last.push_back(reReapCount);
	}

	friend auto operator<<(std::ostream& os, const Stats& s) -> std::ostream&;
	friend auto operator<<(std::ostream& os, const Stats::CompletionType& t) -> std::ostream&;
};
}	 // namespace emper::io
