// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#include "io.hpp"

#include <netdb.h>
#include <sys/socket.h>

#include <cerrno>
#include <cstring>
#include <filesystem>
#include <ostream>
#include <stdexcept>

#include "Common.hpp"
#include "CountingPrivateSemaphore.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Fiber.hpp"
#include "Runtime.hpp"
#include "SynchronizedFiber.hpp"

namespace fs = std::filesystem;

namespace emper::io {

static void recursive_directory_walk_helper(
		CPS& cps, const fs::path& dirpath,
		std::function<bool(const fs::directory_entry& dirent)>& filter,
		std::function<void(fs::directory_entry dirent)>& fn) {
	if constexpr (!emper::HAS_FS_PATH) {
		throw std::logic_error("recursive_directory_walk need c++17 std::filesystem::path");
	}

	Runtime* runtime = Runtime::getRuntime();
	for (const auto& p : fs::directory_iterator(dirpath)) {
		if (filter(p)) {
			auto* fnFiber = SynchronizedFiber::from([=] { fn(p); }, cps);
			runtime->schedule(*fnFiber);
		}

		if (p.is_directory()) {
			auto* subDirFiber = SynchronizedFiber::from(
					[&, p = p] { recursive_directory_walk_helper(cps, p, filter, fn); }, cps);
			runtime->schedule(*subDirFiber);
		}
	}
}

void recursive_directory_walk(const fs::path& dirpath,
															std::function<bool(const fs::directory_entry& dirent)> filter,
															std::function<void(fs::directory_entry dirent)> fn) {
	CPS cps;
	recursive_directory_walk_helper(cps, dirpath, filter, fn);
	cps.wait();
}

auto tcp_listener(std::string& host, std::string& port, const std::function<void(int)>& handler,
									int backlog) -> Fiber* {
	int listen_socket;
	struct addrinfo* servaddr;

	int err = getaddrinfo(host.c_str(), port.c_str(), nullptr, &servaddr);
	if (err) {
		if (err == EAI_SYSTEM) {
			LOGE("getaddrinfo failed: " << strerror(errno));
			return nullptr;
		}

		LOGE("error in getaddrinfo: " << gai_strerror(err));
		return nullptr;
	}

	// socket creation and verification
	listen_socket = socket(servaddr->ai_family, servaddr->ai_socktype, servaddr->ai_protocol);
	if (listen_socket == -1) {
		LOGE("listen socket creation failed: " << strerror(errno));
		return nullptr;
	}

	int reuseaddr = 1;
	if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, sizeof(reuseaddr)) == -1) {
		LOGE("setsockopt failed: " << strerror(errno));
		return nullptr;
	}

	if (bind(listen_socket, servaddr->ai_addr, servaddr->ai_addrlen) == -1) {
		LOGE("bind failed: " << strerror(errno));
		return nullptr;
	}

	if (listen(listen_socket, backlog) != 0) {
		LOGE("listen failed: " << strerror(errno));
		return nullptr;
	}

	Runtime* runtime = Runtime::getRuntime();

	return Fiber::from([=] {
		for (;;) {
			int client_fd = acceptAndWait(listen_socket, nullptr, nullptr);
			if (unlikely(client_fd < 0)) {
				LOGE("accept failed: " << strerror(errno));
				return;
			}

			Fiber* handler_fiber = Fiber::from([=] { handler(client_fd); });
			runtime->schedule(*handler_fiber);
		}
	});
}
}	 // namespace emper::io
