// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus
#pragma once

#include <atomic>
#include <cstdint>

#include "Common.hpp"

namespace adt {

/**
 * @file
 *
 * A CL queue used for work stealing.
 *
 * Implements the partially concurrent queue for the work stealing
 * scheduling algorithm as defined in "Dynamic Circular Work-Stealing Deque" by Chace and Lev. The
 * worker uses pushBottom to push new items on the stack, and
 * retrieves items via popBottom. Eventually, when its own queue is
 * empty, it will use popTop to retrieve items from a different
 * Worker's queue.
 *
 * Only pushBottom/popBottom and popTop may be used concurrently,
 * i.e. concurrent use of pushBottom with popBottom is not allowed.
 *
 * Unlike the ABP queue, the CL queue uses a circular array and
 * only-incrementing uint64_t bottom and top fields. According to the
 * authors "A 64-bit integer is large enough to accommodate 64 years
 * of pushes, pops and steals executing at a rate of 4 billtion
 * operations per second", so overflows should be no problem.
 */
template <typename PAYLOAD, const uintptr_t CAPACITY>
class WsClQueue {
	ALIGN_TO_CACHE_LINE std::atomic<uint64_t> bottom;
	ALIGN_TO_CACHE_LINE std::atomic<uint64_t> top;
	// NOLINTNEXTLINE(modernize-avoid-c-arrays)
	PAYLOAD queue[CAPACITY];

 public:
	// 'bottom' and 'top' are initialized to '1', instead of '0' as
	// it's done in the "Dynamic Circular Work-Stealing Deque" paper
	// because popBottom will decrement bottom, which will result in
	// an underflow if bottom is '0'. The paper's queue uses Java
	// 'long' for bottom and top and is thus safe since it's signed.
	WsClQueue() : bottom(1), top(1) {}
	// TODO: Decide what to do regarding the following suppressed lint.
	// NOLINTNEXTLINE(readability-avoid-const-params-in-decls)
	auto pushBottom(const PAYLOAD item) -> bool;
	auto popTop(PAYLOAD *item) -> bool;
	auto popBottom(PAYLOAD *item) -> bool;
	[[nodiscard]] auto isFull() const -> bool;
	[[nodiscard]] auto isEmpty() const -> bool;
	void print() const;
	[[nodiscard]] inline auto capacity() const -> uintptr_t { return CAPACITY; }
	[[nodiscard]] inline auto usedSlots() const -> uint64_t { return bottom - top; }
	[[nodiscard]] inline auto freeSlots() const -> uintptr_t { return capacity() - usedSlots(); }
};

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClQueue<PAYLOAD, CAPACITY>::pushBottom(const PAYLOAD item) -> bool {
	if (isFull()) return false;

	queue[bottom % CAPACITY] = item;

	// Write fence / memory barrier
	atomic_thread_fence(std::memory_order_release);

	++bottom;
	return true;
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClQueue<PAYLOAD, CAPACITY>::popTop(PAYLOAD *item) -> bool {
start:
	uint64_t oldTop = top;
	if (bottom <= oldTop) return false;

	// Read fence / memory barrier
	atomic_thread_fence(std::memory_order_acquire);

	*item = queue[oldTop % CAPACITY];
	uint64_t newTop = oldTop + 1;

	// Read fence / memory barrier
	atomic_thread_fence(std::memory_order_acquire);

	// If cas fails, then this popTop() lost the race against another
	// popTop().
	if (!top.compare_exchange_weak(
					oldTop, newTop))	//, std::memory_order_release, std::memory_order_relaxed))
		goto start;

	return true;
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClQueue<PAYLOAD, CAPACITY>::popBottom(PAYLOAD *item) -> bool {
	uint64_t localBottom = --bottom;
	uint64_t localTop = top;
	if (localTop > localBottom) {
		// The queue is empty.
		bottom = localTop;
		return false;
	}

	*item = queue[localBottom % CAPACITY];
	if (localBottom > localTop) return true;

	//	bool res = top.compare_exchange_weak(localTop, localTop + 1, std::memory_order_release,
	// std::memory_order_relaxed);
	bool res = top.compare_exchange_weak(localTop, localTop + 1);
	// Either a popTop() removed the element ('res' is false) or we
	// removed the element ('res' is true), but we need to increment
	// the 'bottom' value, since the element bottom pointed at is now
	// gone. N.B. bottom does point to the next free slot, the actual
	// element we remove is bottom-1.
	bottom = localBottom + 1;
	return res;
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClQueue<PAYLOAD, CAPACITY>::isFull() const -> bool {
	return usedSlots() >= CAPACITY;
}

template <typename PAYLOAD, const uintptr_t CAPACITY>
auto WsClQueue<PAYLOAD, CAPACITY>::isEmpty() const -> bool {
	return top >= bottom;
}

}	 // namespace adt
