// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <atomic>
#include <climits>

#include "Common.hpp"

namespace emper::lib::sync {

class CountingTryLock {
 private:
	// The lower 32-bit are used as the actual lock
	// The higher 32-bit are used as a counter which is incremented if
	// try_lock_or_increment fails to acquire the lock.

	// Layout of our lock union:
	// |--------------  countingLock  --------------|
	//                     |------- counter  -------|
	// | lock |
	// 0    sizeof(std::atomic<bool>)              64
	// 0                COUNTER_SHIFT              64
	union {
		std::atomic<uint64_t> countingLock;
		std::atomic<bool> lock;
	};

	static const int COUNTER_SHIFT = 32;
	static_assert(sizeof(std::atomic<bool>) * CHAR_BIT < COUNTER_SHIFT);

	static const uint64_t LOCKED = 1;
	static const uint64_t UNLOCKED = 0;

 public:
	CountingTryLock(bool locked = false) : countingLock(locked ? LOCKED : UNLOCKED) {}

	[[nodiscard]] auto try_lock() -> bool { return !lock.exchange(true, std::memory_order_acquire); }

	[[nodiscard]] auto try_lock_or_increment() -> bool {
		uint64_t oldVal, newVal;
		oldVal = countingLock.load(std::memory_order_relaxed);
		for (;;) {
			// currently unlocked -> try to lock
			if (oldVal == UNLOCKED) {
				newVal = LOCKED;
				// currently locked -> increment the counter
			} else {
				newVal = oldVal + (1L << COUNTER_SHIFT);
			}

			if (countingLock.compare_exchange_weak(oldVal, newVal, std::memory_order_acquire,
																						 std::memory_order_relaxed)) {
				break;
			}
		}

		return oldVal == UNLOCKED;
	}

	// release the lock, zero and return the counter
	auto unlock() -> uint32_t {
		uint64_t oldVal = countingLock.exchange(UNLOCKED, std::memory_order_release);
		return static_cast<uint32_t>(oldVal >> COUNTER_SHIFT);
	}
};
}	 // namespace emper::lib::sync
