// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#include "sleep_strategy/PipeSleepStrategy.hpp"

#include <liburing.h>

#include <atomic>
#include <cassert>
#include <string>

#include "CallerEnvironment.hpp"
#include "Emper.hpp"
#include "Worker.hpp"
#include "io/IoContext.hpp"
#include "lib/TaggedPtr.hpp"

using emper::io::IoContext;
using emper::lib::TaggedPtr;

namespace emper::sleep_strategy {

template <CallerEnvironment callerEnvironment>
[[nodiscard]] auto PipeSleepStrategy::createHint() -> TaggedPtr {
	if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
		return TaggedPtr((uintptr_t)Worker::getCurrentWorkerId(),
										 static_cast<uint16_t>(IoContext::PointerTags::NewWorkWsq));
	} else {
		return TaggedPtr((uintptr_t) nullptr, static_cast<uint16_t>(IoContext::PointerTags::NewWorkAq));
	}
}

template auto PipeSleepStrategy::createHint<CallerEnvironment::EMPER>() -> TaggedPtr;
template auto PipeSleepStrategy::createHint<CallerEnvironment::ANYWHERE>() -> TaggedPtr;

void PipeSleepStrategy::sleep() {
	IoContext& io = *IoContext::getWorkerIo();

	const bool mustPrepareSqe = !io.waitInflight;
	if (mustPrepareSqe) {
		// increment the sleeper count if it was negative we should skip sleeping
		int64_t sleeping = sleepers.fetch_add(1, std::memory_order_acquire);
		if (sleeping < 0) {
			LOGD("observed sleeper count as: " << sleeping << " -> skip sleeping");
			return;
		}

		assert(sleeping <= workerCount);

		struct io_uring_sqe* sqe = io_uring_get_sqe(&io.ring);
		assert(sqe);

		// We read directly into the workers dispatchHint
		io_uring_prep_read(sqe, sleepFd, &io.worker->dispatchHint, sizeof(io.worker->dispatchHint), 0);

		// Mark the sqe as a new work notification to reset the waitInflight flag when reaping the
		// resulting cqe
		io_uring_sqe_set_data(sqe, TaggedPtr((uintptr_t) nullptr,
																				 static_cast<uint16_t>(IoContext::PointerTags::NewWorkAq)));

		if constexpr (emper::DEBUG) {
			io.reqs_in_uring++;
		}

		LOGD("prepared sleepFd read and set sleepers count to: " << sleeping + 1);

		// Before going to sleep prevent the completer from reaping completions
		// on our IoContext otherwise we would neither safely wakeup nor reset
		// the waitInflight flag.
		io.waitInflight.store(true, std::memory_order_acquire);
	}

	// Wait for IO completions
	LOGD("sleep until IO completions occur");
	io.submitPreparedSqesAndWait<CallerEnvironment::EMPER>(1);
	LOGD("wakeup due to available IO completions");

	statsIncr(stats->wakeup);
}

auto operator<<(std::ostream& os, const PipeSleepStrategy::Stats& s) -> std::ostream& {
	os << "PipeSleepStrategy Stats:" << std::endl;
	os << "total-onNewWork: " << std::to_string(s.onNewWork) << std::endl;
	os << "total-notify: " << std::to_string(s.notify) << std::endl;
	os << "total-notifications: " << std::to_string(s.notifications) << std::endl;
	os << "total-skip: " << std::to_string(s.skip) << std::endl;
	os << "total-wakeup: " << std::to_string(s.wakeup) << std::endl;
	os << "total-wakeup-due-to-notify: " << std::to_string(s.wakeupDueToNotify) << std::endl;
	return os;
}
}	 // namespace emper::sleep_strategy
