// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <unistd.h>

#include <algorithm>
#include <atomic>
#include <cstddef>
#include <cstdint>
#include <iostream>
#include <memory>
#include <stdexcept>
#include <type_traits>
#include <vector>

#include "CallerEnvironment.hpp"
#include "Common.hpp"
#include "Debug.hpp"
#include "Emper.hpp"
#include "Worker.hpp"
#include "emper-common.h"
#include "lib/TaggedPtr.hpp"
#include "sleep_strategy/AbstractWorkerSleepStrategy.hpp"

using emper::lib::TaggedPtr;

namespace emper::io {
class IoContext;
}

namespace emper::sleep_strategy {

/**
 * @brief A pipe(2) based sleep strategy combining the IO subsystem with the sleep strategy
 *
 * implement a pipe based sleep strategy using the IO subsystem
 *
 * Design goals
 * ============
 *
 * * Wakeup either on external newWork notifications or on local IO completions
 *   -> Sleep strategy is sound without the IO completer
 * * Do as less as possible in a system saturated with work
 * * Pass a hint where to find new work to suspended workers
 *
 * Algorithm
 * =========
 *
 *```
 * Data:
 *     Global:
 *             hint pipe
 *             sleepers count
 *     Per worker:
 *             dispatch hint buffer
 *             in flight flag
 *
 * Sleep:
 *     if we have no sleep request in flight
 *             Atomic increment sleep count
 *             Remember that we are sleeping
 *             Prepare read cqe from the hint pipe to dispatch hint buffer
 *     Prevent the completer from reaping completions on this worker IoContext
 *     Wait until IO completions occurred
 *
 * NotifyEmper(n):
 *    if observed sleepers <= 0
 *            return
 *
 *    // Determine how many we are responsible to wake
 *    do
 *            toWakeup = min(observed sleepers, n)
 *    while (!CAS(sleepers, toWakeup))
 *
 *    write toWakeup hints to the hint pipe
 *
 * NotifyAnywhere(n):
 *    // Ensure all n notifications take effect
 *    while (!CAS(sleepers, observed sleepers - n))
 *        if observed sleeping <= -n
 *            return
 *
 *    toWakeup = min(observed sleeping, n)
 *    write toWakeup hints to the hint pipe
 *
 * onNewWorkCompletion:
 *     reset in flight flag
 *     allow completer to reap completions on this IoContext
 *```
 *
 * Notes
 * =====
 *
 * * We must decrement the sleepers count on the notifier side to
 *   prevent multiple notifiers to observe all the same amount of sleepers,
 *   trying to wake up the same sleepers by writing to the pipe and jamming it up
 *   with unconsumed hints and thus blocking in the notify write resulting
 *   in a deadlock.
 * * The CAS loops on the notifier side are needed because decrementing
 *   and incrementing the excess is racy: Two notifier can observe the
 *   sum of both their excess decrement and increment to much resulting in a
 *   broken counter.
 * * Add the dispatch hint code in AbstractWorkStealingScheduler::nextFiber.
 *   This allows workers to check the dispatch hint after there
 *   where no local work to execute.
 *   This is a trade-off where we trade slower wakeup - a just awoken worker
 *   will check for local work - against a faster dispatch hot path when
 *   we have work to do in our local WSQ.
 * * The completer tread must not reap completions on the IoContexts of
 *   sleeping workers because this introduces a race for cqes and a possible
 *   lost wakeup if the completer consumes the completions before the worker
 *   is actually waiting for them.
 * * When notifying sleeping workers from anywhere we must ensure that all
 *   notifications take effect. This is needed for example when terminating
 *   the runtime to prevent sleep attempt from worker thread which are
 *   about to sleep but have not incremented the sleeper count yet.
 *   We achieve this by always decrementing the sleeper count by the notification
 *   count.
 */
class PipeSleepStrategy : AbstractWorkerSleepStrategy<PipeSleepStrategy>,
													public Logger<LogSubsystem::SLEEP_S> {
	friend class emper::io::IoContext;

	class Stats {
	 public:
		std::atomic<size_t> onNewWork = 0;
		std::atomic<size_t> notify = 0;
		std::atomic<size_t> notifications = 0;
		std::atomic<size_t> skip = 0;
		std::atomic<size_t> wakeup = 0;
		std::atomic<size_t> wakeupDueToNotify = 0;
	};

	std::unique_ptr<Stats> stats;

	workerid_t workerCount;
	int sleepFd;
	int notifyFd;

	// Make sure the shared counter lives in an exlusive cache line
	std::aligned_storage<64, 64>::type sleepers_mem;
	std::atomic<int64_t>& sleepers = *reinterpret_cast<std::atomic<int64_t>*>(&sleepers_mem);

	template <CallerEnvironment callerEnvironment>
	[[nodiscard]] auto createHint() -> TaggedPtr;

	void writeNotifications(TaggedPtr hint, int64_t count) {
		statsAdd(stats->notifications, count);

		std::vector<void*> hints(count, hint);

		ssize_t res = write(notifyFd, hints.data(), sizeof(void*) * hints.size());
		if (unlikely(res) < 0) {
			DIE_MSG_ERRNO("writing to notifyFd failed");
		}
	}

	void notifyFromEmper(int64_t& sleeping, TaggedPtr hint, int64_t count) {
		int64_t toWakeup;

		// We need to decrement the sleeper count on the notification side
		// to prevent overcommiting notifications, filling up the pipe which results
		// in notify calls to block on write.
		// And the CAS loop is needed because we don't want to decrement the sleepers
		// count to much.

		// Decrement the sleeper count for each sleeper we are going to wakeup
		do {
			// No one is sleeping -> we don't need to notify
			if (sleeping <= 0) {
				return;
			}

			// Count of sleepers we notify
			toWakeup = std::min(sleeping, count);
		} while (!sleepers.compare_exchange_weak(sleeping, sleeping - toWakeup,
																						 std::memory_order_release, std::memory_order_acquire));

		// increment the notify stat if we actually have to notify someone
		statsIncr(stats->notify);
		writeNotifications(hint, toWakeup);

		LOGD("notifyFromEmper written "
				 << toWakeup << " notifications and set sleepers count to: " << sleeping - toWakeup);
	}

	void notifyFromAnywhere(int64_t& sleeping, TaggedPtr hint, int64_t count) {
		// If we observe nobody sleeping we need to prevent sleep locks when
		// everybody is about to sleep but none has incremented the sleepers count yet.
		// We prevent the next sleep from blocking by setting the sleeper count to
		// the negative amount we want to prevent from sleeping.

		// More general if we notify from anywhere we need all notifications to take
		// effect. Only notifying the sleepers we observe may not be enough.
		// For example when terminating the runtime we notify all workers if we
		// only wakeup the ones we observe sleeping we may never terminate because
		// after we notified all we observed new one went to sleep which will never
		// be notified.
		// To prevent this we make sure that all count notifications take effect
		// by making the sleepers count negative and thus preventing at least count
		// sleep attempts.

		// Always increment the notify count because notifications from anywhere always
		// take effect
		statsIncr(stats->notify);

		do {
			// We already prevent enough sleep attempts
			if (sleeping <= -count) {
				LOGD("notifyFromAnywhere sleeper count already preventing enough sleep attempts");
				return;
			}
		} while (!sleepers.compare_exchange_weak(sleeping, sleeping - count, std::memory_order_release,
																						 std::memory_order_acquire));

		int64_t toWakeup = std::min(sleeping, count);
		if (sleeping > 0) {
			writeNotifications(hint, toWakeup);
		}

		LOGD("notifyFromAnywhere written "
				 << toWakeup << " notifications and set sleepers count to: " << sleeping - count);
	}

	template <CallerEnvironment callerEnvironment>
	void notify(TaggedPtr hint, uint32_t count) {
		// The hint must be != nullptr so sleep() knows when to prepare and submit
		// a sleepFd read sqe.

		statsIncr(stats->onNewWork);

		int64_t sleeping = getSleeping();
		auto signedCount = static_cast<int64_t>(count);

		if constexpr (callerEnvironment == CallerEnvironment::ANYWHERE) {
			notifyFromAnywhere(sleeping, hint, signedCount);
		} else {
			notifyFromEmper(sleeping, hint, signedCount);
		}
	}

 public:
	PipeSleepStrategy(workerid_t workerCount) : workerCount(workerCount) {
		LOGD("init pipe sleep startegy");
		sleepers.store(0, std::memory_order_relaxed);

		// NOLINTNEXTLINE(modernize-avoid-c-arrays)
		int fds[2];
		if (pipe(fds)) {
			DIE_MSG_ERRNO("pipe failed");
		}
		sleepFd = fds[0];
		notifyFd = fds[1];

		if constexpr (emper::STATS) {
			stats = std::make_unique<Stats>();
		}
	}

	void printStats() { std::cout << *stats; }

	[[nodiscard]] inline auto getSleeping() const -> long { return sleepers.load(); }

	template <CallerEnvironment callerEnvironment>
	inline void notifyOne() {
		notifyMany<callerEnvironment>(1);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyMany(unsigned count) {
		if (count > 1) LOGD("notify(hint, " << count << ")");
		notify<callerEnvironment>(createHint<callerEnvironment>(), count);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyAll() {
		notifyMany<callerEnvironment>(workerCount);
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifySpecific(workerid_t /* workerId */) {
		throw std::logic_error("Not implemented");
		// // TODO: get pid of specific worker
		// pid_t specific = 0;
		// notify<callerEnvironment>(createHint<callerEnvironment>(), specific);
	}

	void sleep();

	friend auto operator<<(std::ostream& os, const PipeSleepStrategy::Stats& s) -> std::ostream&;
};

auto operator<<(std::ostream& os, const PipeSleepStrategy::Stats& s) -> std::ostream&;
}	 // namespace emper::sleep_strategy
