// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021 Florian Fischer
#pragma once

#include <atomic>

#include "CallerEnvironment.hpp"
#include "Worker.hpp"
#include "emper-common.h"
#include "emper-config.h"
#include "sleep_strategy/AbstractWorkerSleepStrategy.hpp"

#ifdef EMPER_LOCKED_WAKEUP_SEMAPHORE
#include "lib/sync/LockedSemaphore.hpp"
#endif

#ifdef EMPER_POSIX_WAKEUP_SEMAPHORE
#include "lib/sync/PosixSemaphore.hpp"
#endif

#ifdef EMPER_FUTEX_WAKEUP_SEMAPHORE
#include "lib/sync/SpuriousFutexSemaphore.hpp"
#endif

namespace emper::sleep_strategy {

// Needed to stop waiting for a worker to change its flag if it may be already
// terminated
auto isRuntimeTerminating() -> bool;

template <class Sem>
class AbstractSemaphoreWorkerSleepStrategy
		: AbstractWorkerSleepStrategy<AbstractSemaphoreWorkerSleepStrategy<Sem>> {
	const workerid_t workerCount;
	Sem wakeupSem;

	// check if the used Semaphore provides a notifySpecific implementation
	static constexpr bool semHasNotifySpecific = requires(Sem s) { s.notifySpecific(0); };

	// should the generic notifySpecific implementation be used
	static constexpr bool useGenericNotifySpecificImpl = !semHasNotifySpecific && needsNotifySpecific;

	// Member used for the generic notifySpecific implementation
	std::atomic<bool>* notifiedFlags;

	// Sleep part of the generic notifySpecific implementation
	inline void genericNotifySpecificSleep() {
		workerid_t workerId = Worker::getCurrentWorkerId();
		auto& notified = notifiedFlags[workerId];
		if (notified.load(std::memory_order_relaxed)) {
			notified.store(false, std::memory_order_relaxed);
			return;
		}

		wakeupSem.wait();

		if (notified.load(std::memory_order_relaxed)) {
			notified.store(false, std::memory_order_relaxed);
		}
	}

	// Currently we don't have a good sempahore based algorithm to notify a specific
	// worker.
	// Just brute forcing (notifyAll and hope) is unfortunatly not possbile.
	//
	// A single worker can consume more tokens from the semaphore than others,
	// especially if notifyAll is implemented as `foreach worker: notify()`.
	// This could theoretically lead to a "lost-wakeup" for the targeted worker.
	//
	// 0. Two workers W1, W2
	//    IO completer with wakeup behaviour
	//    inflight IO operation for W1
	// 1. W1 is at the end of its dispatch loop going to sleep
	// 2. W2 is sleeping
	// 3. Kernel produces a cqe in CQ of W1
	// 4. Completer notifies all workers -> sem has value 2
	// 5. W2 iterates the dispatch loop two times without finding work -> goes to sleep
	// 6. W1 calls sleep() with a sem value of 0 -> goes to sleep though it has a cqe in its CQ
	//
	// The used approach is to keep a notified flag for each worker in the notifiedFlags
	// array. Those flags make it possible to notify a specific worker.
	// We are sure that the specific worker was successfully notified if we observe
	// that its flag was reset.
	// Therefore we can notify all worker in a loop until we observe that the
	// specific worker has changed its flag or the runtime is terminating.
	//
	// Unfortunately this introduces significant overhead in the sleep method which is
	// only necessary if the runtime actually wants to notify a specific worker.
	// This should be configured at compile time using a boolean template parameter of the
	// WorkerWakeupStrategy (<bool needsNotifySpecific>).
	//
	// For now we hardcode a constexpr check for the only condition where we need notifySpecific
	template <CallerEnvironment callerEnvironment>
	inline void genericNotifySpecific(workerid_t workerId) {
		auto& notifiedFlag = notifiedFlags[workerId];
		notifiedFlag.store(true, std::memory_order_relaxed);
		// NotifyAll since we observe that the specific worker has reset its flag
		while (notifiedFlag.load(std::memory_order_relaxed) && !isRuntimeTerminating()) {
			notifyAll<callerEnvironment>();
		}
	}

	template <CallerEnvironment callerEnvironment>
	[[nodiscard]] inline auto mustNotify() -> bool {
		typename Sem::CounterType skipWakeupThreshold;
		if constexpr (callerEnvironment == CallerEnvironment::ANYWHERE) {
			// On external work we always increment the semaphore unless we observe
			// that its value is > workerCount.
			// If we observe semValue > workerCount we are ensured that some worker will iterate
			// its dispatchLoop again and must observe the new work.
			// TODO: Could it be >= workerCount ?
			skipWakeupThreshold = workerCount;
		} else {
			// For work from within emper we skip wakeup if we observe no one sleeping.
			// This is sound because wakeupSleepingWorkers() is called from a active
			// worker which will observe its own new work in its next dispatchLoop before
			// going to sleep.

			// Note that sem_getvalue() is allowed to return 0 if there are
			// waiting workers, hence we need to set the threshold also to
			// 0. This has the disadvantage that we will perform one
			// unnecessary sem_post. If we are sure the wakeupSem
			// implementation does not return 0 with waiters,
			// then the skipWakeupThreshold value should be
			// reviewed and potentially changed to '-1'.
			skipWakeupThreshold = 0;
		}

		auto semValue = wakeupSem.getValue();
		return !(semValue > skipWakeupThreshold);
	}

 public:
	AbstractSemaphoreWorkerSleepStrategy(workerid_t workerCount) : workerCount(workerCount) {
		if constexpr (useGenericNotifySpecificImpl) {
			notifiedFlags = new std::atomic<bool>[workerCount];
		}
	}

	~AbstractSemaphoreWorkerSleepStrategy() {
		if constexpr (useGenericNotifySpecificImpl) {
			delete[] notifiedFlags;
		}
	}

	[[nodiscard]] inline auto getSleeping() const -> long { return wakeupSem.getValue(); }

	template <CallerEnvironment callerEnvironment>
	inline void notifyOne() {
		if (mustNotify<callerEnvironment>()) {
			wakeupSem.notify();
		}
	}

	template <CallerEnvironment callerEnvironment>
	inline void notifyMany(unsigned count) {
		if (mustNotify<callerEnvironment>()) {
			wakeupSem.notify_many(count);
		}
	}

	// notify all workers
	// It is sound to increment the semaphore to much, thus this will only cause
	// workers to iterate the dispatchLoop more often before actually sleeping
	template <CallerEnvironment callerEnvironment>
	inline void notifyAll() {
		if (mustNotify<callerEnvironment>()) {
			wakeupSem.notify_many(workerCount);
		}
	}

	template <CallerEnvironment callerEnvironment>
	void notifySpecific(workerid_t workerId) {
		if constexpr (semHasNotifySpecific) {
			wakeupSem.notifySpecific(workerId);
		} else {
			genericNotifySpecific<callerEnvironment>(workerId);
		}
	}

	inline void sleep() {
		if constexpr (useGenericNotifySpecificImpl) {
			genericNotifySpecificSleep();
		} else {
			wakeupSem.wait();
		}
	}

	void printStats(){};
};

#ifdef EMPER_LOCKED_WAKEUP_SEMAPHORE
using ::emper::lib::sync::LockedSemaphore;
using SemaphoreWorkerSleepStrategy = AbstractSemaphoreWorkerSleepStrategy<LockedSemaphore>;

#elif defined EMPER_POSIX_WAKEUP_SEMAPHORE
using ::emper::lib::sync::PosixSemaphore;
using SemaphoreWorkerSleepStrategy = AbstractSemaphoreWorkerSleepStrategy<PosixSemaphore>;

#elif defined EMPER_FUTEX_WAKEUP_SEMAPHORE
using ::emper::lib::sync::SpuriousFutexSemaphore;
using SemaphoreWorkerSleepStrategy = AbstractSemaphoreWorkerSleepStrategy<SpuriousFutexSemaphore>;

#else
#error Unknown WorkerSleepSemaphore implementation
#endif

}	 // namespace emper::sleep_strategy
