// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#include "LawsScheduler.hpp"

#include <cstdint>

#include "CallerEnvironment.hpp"
#include "Emper.hpp"
#include "LawsStrategy.hpp"	 // IWYU pragma: keep
#include "NextFiberResult.hpp"
#include "Runtime.hpp"
#include "emper-common.h"
#include "strategies/laws/LawsWorkerStats.hpp"

thread_local LawsScheduler::LawsMpscQueue LawsScheduler::priorityQueue;

LawsScheduler::LawsScheduler(Runtime& runtime, RuntimeStrategy& strategy)
		: AbstractWorkStealingScheduler(runtime, strategy) {
	const workerid_t workerCount = runtime.getWorkerCount();
	priorityQueues = new LawsScheduler::LawsMpscQueue*[workerCount];

	auto newWorkerHook = [this](workerid_t workerId) { priorityQueues[workerId] = &priorityQueue; };
	addNewWorkerHook(newWorkerHook);
}

template <CallerEnvironment callerEnvironment>
void LawsScheduler::tryScheduleToPriorityQueue(Fiber& fiber) {
	workeraffinity_t* const affinity_buffer = getAffinityBuffer(fiber);
	if (!affinity_buffer) {
		return;
	}

	workeraffinity_t affinity = *affinity_buffer;

	if constexpr (callerEnvironment == CallerEnvironment::EMPER) {
		workerid_t workerId = Runtime::getWorkerId();
		if (affinity == workerId) {
			return;
		}
	}

	if (affinity == Fiber::NOT_AFFINE) {
		return;
	}

	// We found a fiber to schedule on a remote prority queue.
	increaseRefCount(fiber);
	priorityQueues[affinity]->enqueue(&fiber);

	emper::statsIncr(LawsStrategy::stats.scheduledFibersToPriority);
}

void LawsScheduler::scheduleInternal(Fiber& fiber) {
	tryScheduleToPriorityQueue<CallerEnvironment::EMPER>(fiber);
	scheduleViaWorkStealing(fiber);
}

void LawsScheduler::scheduleFromAnywhereInternal(Fiber& fiber) {
	tryScheduleToPriorityQueue<CallerEnvironment::ANYWHERE>(fiber);
	enqueueInAnywhereQueue(fiber);
	onNewWork<CallerEnvironment::ANYWHERE>();
}

void LawsScheduler::scheduleFromAnywhereInternal(Fiber** fibers, unsigned count) {
	for (unsigned i = 0; i < count; ++i) {
		Fiber& fiber = *fibers[i];
		tryScheduleToPriorityQueue<CallerEnvironment::ANYWHERE>(fiber);
	}
	insertInAnywhereQueue(fibers, count);
	onNewWork<CallerEnvironment::ANYWHERE>();
}

auto LawsScheduler::nextFiber() -> NextFiberResult {
	Fiber* fiber = priorityQueue.dequeue();
	if (fiber != nullptr) {
		return NextFiberResult{
				fiber,
				static_cast<uintptr_t>(LawsStrategy::FiberSource::fromPriority),
		};
	}

	return nextFiberResultViaWorkStealing();
}
